from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    datawidth = 32
    addrwidth = 10
    reduce_size = 4

    myaxi = vthread.AXIM(m, 'myaxi', clk, rst, datawidth)
    ram_a = vthread.RAM(m, 'ram_a', clk, rst, datawidth, addrwidth)
    ram_b = vthread.RAM(m, 'ram_b', clk, rst, datawidth, addrwidth)
    ram_c = vthread.RAM(m, 'ram_c', clk, rst, datawidth, addrwidth)
    ram_d = vthread.RAM(m, 'ram_d', clk, rst, datawidth, addrwidth)

    macstrm = vthread.Stream(m, 'macstream', clk, rst)
    macstrm_a = macstrm.source('a')
    macstrm_b = macstrm.source('b')
    macstrm_const = macstrm.parameter('const')
    macstrm_mul = macstrm_a * macstrm_b
    macstrm_c, macstrm_v = macstrm.ReduceAddValid(macstrm_mul, macstrm_const)
    macstrm.sink(macstrm_c, 'c')
    macstrm.sink(macstrm_v, 'v')

    macstrm2 = vthread.Stream(m, 'macstream2', clk, rst)
    macstrm2_a = macstrm2.source('a')
    macstrm2_b = macstrm2.source('b')
    macstrm2_const = macstrm2.parameter('const')
    macsub = macstrm2.substream(macstrm)
    macsub.to_source('a', macstrm2_a)
    macsub.to_source('b', macstrm2_b)
    macsub.to_parameter('const', macstrm2_const)
    macstrm2_c = macsub.from_sink('c')
    macstrm2_v = macsub.from_sink('v')
    macstrm2.sink(macstrm2_c, 'c')
    macstrm2.sink(macstrm2_v, 'v')

    neststrm = vthread.Stream(m, 'neststream', clk, rst)
    neststrm_a = neststrm.source('a')
    neststrm_b = neststrm.source('b')
    neststrm_const = neststrm.parameter('const')
    neststrm_a += 1
    neststrm_a += 0
    neststrm_b += 1
    macsub = neststrm.substream(macstrm2)
    macsub.to_source('a', neststrm_a)
    macsub.to_source('b', neststrm_b)
    macsub.to_parameter('const', neststrm_const)
    neststrm_c = macsub.from_sink('c')
    neststrm_c += neststrm_a
    neststrm_c += 0
    neststrm_v = macsub.from_sink('v')
    neststrm.sink(neststrm_c, 'c')
    neststrm.sink(neststrm_v, 'v')

    strm = vthread.Stream(m, 'mystream', clk, rst)
    x = strm.source('x')
    y = strm.source('y')
    const = strm.parameter('const')
    sub = strm.substream(neststrm)
    sub.to_source('a', x)
    sub.to_source('b', y)
    sub.to_parameter('const', const)
    z = sub.from_sink('c')
    v = sub.from_sink('v')
    z = z + y
    strm.sink(z, 'z', when=v, when_name='v')

    all_ok = m.TmpReg(initval=0)

    def comp_stream_macstrm(size, offset):
        macstrm.set_source('a', ram_a, offset, size)
        macstrm.set_source('b', ram_b, offset, size)
        macstrm.set_parameter('const', reduce_size)
        macstrm.set_sink('c', ram_c, offset, size)
        macstrm.set_sink('v', ram_d, offset, size)
        macstrm.run()
        macstrm.join()

    def comp_stream_mystrm(size, offset):
        strm.set_source('x', ram_a, offset, size)
        strm.set_source('y', ram_b, offset, size)
        strm.set_parameter('const', reduce_size)
        strm.set_sink('z', ram_c, offset, size // reduce_size)
        strm.run()
        strm.join()

    def comp_sequential_macstrm(size, offset):
        sum = 0
        count = 0
        for i in range(size):
            a = ram_a.read(i + offset)
            b = ram_b.read(i + offset)
            sum += a * b
            count += 1
            ram_c.write(i + offset, sum)
            ram_d.write(i + offset, count == (reduce_size - 1))
            if count == reduce_size:
                sum = 0
                count = 0

    def comp_sequential_mystrm(size, offset):
        sum = 0
        count = 0
        write_offset = offset
        for i in range(size):
            x = ram_a.read(i + offset)
            y = ram_b.read(i + offset)
            sum += (x + 1) * (y + 1)
            val = sum + (x + 1) + y
            count += 1
            if count == reduce_size:
                ram_c.write(write_offset, val)
                write_offset += 1
                sum = 0
                count = 0

    def check(size, offset_stream, offset_seq):
        for i in range(size):
            st = ram_c.read(i + offset_stream)
            sq = ram_c.read(i + offset_seq)
            if vthread.verilog.NotEql(st, sq):
                all_ok.value = False
                print(i, st, sq)
        if all_ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

    def comp(size):
        all_ok.value = True

        # stream
        offset = 0
        myaxi.dma_read(ram_a, offset, 0, size)
        myaxi.dma_read(ram_b, offset, 0, size)
        comp_stream_macstrm(size, offset)
        myaxi.dma_write(ram_c, offset, 1024, size)

        # sequential
        offset = size
        myaxi.dma_read(ram_a, offset, 0, size)
        myaxi.dma_read(ram_b, offset, 0, size)
        comp_sequential_macstrm(size, offset)
        myaxi.dma_write(ram_c, offset, 1024 * 2, size)

        # verification
        print('# macstream')
        check(size, 0, offset)

        # stream
        offset = 0
        myaxi.dma_read(ram_a, offset, 0, size)
        myaxi.dma_read(ram_b, offset, 0, size)
        comp_stream_mystrm(size, offset)
        myaxi.dma_write(ram_c, offset, 1024, size // reduce_size)

        # sequential
        offset = size
        myaxi.dma_read(ram_a, offset, 0, size)
        myaxi.dma_read(ram_b, offset, 0, size)
        comp_sequential_mystrm(size, offset)
        myaxi.dma_write(ram_c, offset, 1024 * 2, size // reduce_size)

        # verification
        print('# mystream')
        check(size // reduce_size, 0, offset)

        vthread.finish()

    th = vthread.Thread(m, 'th_comp', clk, rst, comp)
    fsm = th.start(16)

    return m


def mkTest(memimg_name=None):
    m = Module('test')

    # target instance
    led = mkLed()

    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst, memimg_name=memimg_name)
    memory.connect(ports, 'myaxi')

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    #simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)
    lines = rslt.splitlines()
    if simtype == 'verilator' and lines[-1].startswith('-'):
        rslt = '\n'.join(lines[:-1])
    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import functools

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi
import veriloggen.types.util as util


def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    datawidth = 32
    addrwidth = 10
    myaxi = vthread.AXIM(m, 'myaxi', clk, rst, datawidth)

    saxi_length = 5
    saxi = vthread.AXISLiteRegister(m, 'saxi', clk, rst, datawidth=datawidth, length=saxi_length)

    ram_src = vthread.RAM(m, 'ram_src', clk, rst, datawidth, addrwidth)
    ram_dummy_src = vthread.RAM(m, 'ram_dummy_src', clk, rst, datawidth, addrwidth)
    ram_dst = vthread.RAM(m, 'ram_dst', clk, rst, datawidth, addrwidth)

    strm = vthread.Stream(m, 'mystream', clk, rst)
    dummy_src = strm.source('dummy_src')
    c = strm.Counter(initval=0, size=4)
    x = strm.Counter(initval=0, size=8, enable=(c == 3))
    y = strm.Counter(initval=0, size=8, enable=((c == 3) & (x == 7)))

    shift_cond = (x & 1 == 0) & ((y & 1) == 0)
    rotate_cond1 = (((((x & 1) == 0) & ((y & 1) == 0)) == 0) & (((x & 1) == 0) == 0))
    rotate_cond2 = (((((x & 1) == 0) & ((y & 1) == 0)) == 0) & ((x & 1) == 0))
    read_cond = shift_cond
    addrcounter = strm.Counter(initval=0, enable=read_cond)
    src = strm.read_RAM('ram_src', addr=addrcounter, when=read_cond, datawidth=datawidth)
    counter = strm.Counter(initval=0)
    width = strm.parameter('width')
    height = strm.parameter('height')

    linebuf = strm.LineBuffer(shape=(1, 1, 1), memlens=[4, 13],
                              head_initvals=[0, 0], tail_initvals=[3, 12],
                              data=src, shift_cond=shift_cond,
                              rotate_conds=[rotate_cond1, rotate_cond2])
    dst = linebuf.get_window(0)

    strm.sink(dst, 'dst')

    # add a stall condition
    count = m.Reg('count', 4, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        count.inc()
    )

    util.add_disable_cond(strm.oready, 1, count == 0)

    def comp_stream(channel, width, height, offset):
        strm.set_source('dummy_src', ram_dummy_src, offset, channel * width * height * 2 * 2)
        strm.set_read_RAM('ram_src', ram_src)
        strm.set_sink('dst', ram_dst, offset, channel * width * height * 2 * 2)
        strm.set_parameter('width', width)
        strm.set_parameter('height', height)
        strm.run()
        strm.join()

    def comp_sequential(channel, width, height, roffset, woffset):
        for yy in range(height * 2):
            for xx in range(width * 2):
                for c in range(channel):
                    # f(c, x, y) = in(c, x/2, y/2);
                    src_i = (xx // 2) * channel + (yy // 2) * width * channel + c
                    dst_i = xx * channel + yy * width * 2 * channel + c
                    val = ram_src.read(roffset + src_i)
                    ram_dst.write(woffset + dst_i, val)

    def check(offset_stream, offset_seq, size):
        all_ok = True
        for i in range(size):
            st = ram_dst.read(offset_stream + i)
            sq = ram_dst.read(offset_seq + i)
            if vthread.verilog.NotEql(st, sq):
                all_ok = False
        if all_ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

    def comp():
        saxi.write(addr=1, value=0)
        saxi.wait_flag(0, value=1, resetvalue=0)
        channel = saxi.read(2)
        width = saxi.read(3)
        height = saxi.read(4)
        insize = channel * width * height
        outsize = channel * width * 2 * height * 2

        roffset = 0
        woffset = 0
        myaxi.dma_read(ram_src, roffset, 0, insize)
        comp_stream(channel, width, height, roffset)
        myaxi.dma_write(ram_dst, woffset, 1024, outsize)

        roffset = insize
        woffset = outsize
        myaxi.dma_read(ram_src, roffset, 0, insize)
        comp_sequential(channel, width, height, roffset, woffset)
        myaxi.dma_write(ram_dst, woffset, 2 * 1024, outsize)

        check(0, woffset, outsize)
        saxi.write(addr=1, value=1)

    th = vthread.Thread(m, 'th_comp', clk, rst, comp)
    fsm = th.start()

    return m


def mkTest(memimg_name=None):
    m = Module('test')

    # target instance
    led = mkLed()

    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst, memimg_name=memimg_name)
    memory.connect(ports, 'myaxi')
    maxi = vthread.AXIMLite(m, 'maxi', clk, rst, noio=True)
    maxi.connect(ports, 'saxi')

    def ctrl():
        channel, width, height = [4, 4, 4]

        awaddr = 2 * 4
        maxi.write(awaddr, channel)

        awaddr = 3 * 4
        maxi.write(awaddr, width)

        awaddr = 4 * 4
        maxi.write(awaddr, height)

        awaddr = 0 * 4
        maxi.write(awaddr, 1)

        araddr = 1 * 4
        v = maxi.read(araddr)
        while v == 0:
            v = maxi.read(araddr)
        vthread.finish()

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    # simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)
    lines = rslt.splitlines()
    if simtype == 'verilator' and lines[-1].startswith('-'):
        rslt = '\n'.join(lines[:-1])
    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import veriloggen.types.util as util
import veriloggen.types.axi as axi
from veriloggen.fsm.fsm import FSM

from .ttypes import _MutexFunction


class AXIS(axi.AxiSlave, _MutexFunction):
    __intrinsics__ = _MutexFunction.__intrinsics__

    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=32,
                 waddr_id_width=0, wdata_id_width=0, wresp_id_width=0,
                 raddr_id_width=0, rdata_id_width=0,
                 waddr_user_width=2, wdata_user_width=0, wresp_user_width=0,
                 raddr_user_width=2, rdata_user_width=0,
                 wresp_user_mode=axi.xUSER_DEFAULT,
                 rdata_user_mode=axi.xUSER_DEFAULT,
                 noio=False):

        axi.AxiSlave.__init__(self, m, name, clk, rst, datawidth, addrwidth,
                              waddr_id_width, wdata_id_width, wresp_id_width,
                              raddr_id_width, rdata_id_width,
                              waddr_user_width, wdata_user_width, wresp_user_width,
                              raddr_user_width, rdata_user_width,
                              wresp_user_mode,
                              rdata_user_mode,
                              noio)
        self.mutex = None


class AXISLite(axi.AxiLiteSlave, _MutexFunction):
    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=32,
                 noio=False):

        axi.AxiLiteSlave.__init__(self, m, name, clk, rst, datawidth, addrwidth,
                                  noio)
        self.mutex = None


class AXISRegister(AXIS):
    __intrinsics__ = ('read', 'write', 'write_flag', 'wait',
                      'wait_flag') + _MutexFunction.__intrinsics__

    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=32,
                 waddr_id_width=0, wdata_id_width=0, wresp_id_width=0,
                 raddr_id_width=0, rdata_id_width=0,
                 waddr_user_width=2, wdata_user_width=0, wresp_user_width=0,
                 raddr_user_width=2, rdata_user_width=0,
                 wresp_user_mode=axi.xUSER_DEFAULT,
                 rdata_user_mode=axi.xUSER_DEFAULT,
                 noio=False, length=4, fsm_as_module=False):

        AXIS.__init__(self, m, name, clk, rst, datawidth, addrwidth,
                      waddr_id_width, wdata_id_width, wresp_id_width,
                      raddr_id_width, rdata_id_width,
                      waddr_user_width, wdata_user_width, wresp_user_width,
                      raddr_user_width, rdata_user_width,
                      wresp_user_mode,
                      rdata_user_mode,
                      noio)

        self.fsm_as_module = fsm_as_module

        if not isinstance(length, int):
            raise TypeError("length must be 'int', not '%s'" %
                            str(type(length)))

        self.register = [self.m.Reg('_'.join(['', self.name, 'register', '%d' % i]),
                                    width=self.datawidth, initval=0, signed=True)
                         for i in range(length)]
        self.flag = [self.m.Reg('_'.join(['', self.name, 'flag', '%d' % i]), initval=0)
                     for i in range(length)]
        self.resetval = [self.m.Reg('_'.join(['', self.name, 'resetval', '%d' % i]),
                                    width=self.datawidth, initval=0, signed=True)
                         for i in range(length)]
        self.length = length
        self.maskwidth = self.m.Localparam('_'.join(['', self.name, 'maskwidth']),
                                           util.log2(length))
        self.mask = self.m.Localparam('_'.join(['', self.name, 'mask']),
                                      vtypes.Repeat(vtypes.Int(1, 1), self.maskwidth))
        self.shift = self.m.Localparam('_'.join(['', self.name, 'shift']),
                                       util.log2(self.datawidth // 8))

        self._set_register_full_fsm()

    def _set_register_full_fsm(self):
        fsm = FSM(self.m, '_'.join(['', self.name, 'register_fsm']),
                  self.clk, self.rst, as_module=self.fsm_as_module)

        # request
        addr, counter, readvalid, writevalid = self.pull_request_counter(cond=fsm)

        maskaddr = self.m.TmpReg(self.maskwidth)
        fsm.If(vtypes.Ors(readvalid, writevalid))(
            maskaddr((addr >> self.shift) & self.mask),
        )

        init_state = fsm.current

        # read
        read_state = fsm.current + 1
        fsm.If(readvalid).goto_from(init_state, read_state)
        fsm.set_index(read_state)

        rdata = self.m.TmpWire(self.datawidth, signed=True)
        pat = [(maskaddr == i, r) for i, r in enumerate(self.register)]
        pat.append((None, vtypes.IntX()))
        rval = vtypes.PatternMux(pat)
        rdata.assign(rval)

        flag = self.m.TmpWire()
        pat = [(maskaddr == i, r) for i, r in enumerate(self.flag)]
        pat.append((None, vtypes.IntX()))
        rval = vtypes.PatternMux(pat)
        flag.assign(rval)

        resetval = self.m.TmpWire(self.datawidth, signed=True)
        pat = [(maskaddr == i, r) for i, r in enumerate(self.resetval)]
        pat.append((None, vtypes.IntX()))
        rval = vtypes.PatternMux(pat)
        resetval.assign(rval)

        ack, valid, last = self.push_read_data(rdata, counter, cond=fsm)

        # flag reset
        state_cond = fsm.state == fsm.current
        for i, r in enumerate(self.register):
            self.seq.If(state_cond, ack, flag, maskaddr == i)(
                self.register[i](resetval),
                self.flag[i](0)
            )

        fsm.If(ack)(
            maskaddr.inc()
        )
        fsm.If(valid, last).goto_init()

        # write
        write_state = fsm.current + 1
        fsm.If(writevalid).goto_from(init_state, write_state)
        fsm.set_index(write_state)

        data, mask, valid, last = self.pull_write_data(counter, cond=fsm)

        state_cond = fsm.state == fsm.current
        for i, r in enumerate(self.register):
            self.seq.If(state_cond, valid, maskaddr == i)(
                self.register[i](data)
            )

        fsm.If(valid)(
            maskaddr.inc()
        )
        fsm.If(valid, last).goto_init()

    def read(self, fsm, addr):
        if isinstance(addr, int):
            rval = self.register[addr]
        elif isinstance(addr, vtypes.Int):
            rval = self.register[addr.value]
        else:
            pat = [(addr == i, r) for i, r in enumerate(self.register)]
            pat.append((None, vtypes.IntX()))
            rval = vtypes.PatternMux(pat)
        return rval

    def write(self, fsm, addr, value):
        state_cond = fsm.state == fsm.current
        for i, r in enumerate(self.register):
            self.seq.If(state_cond, addr == i)(
                self.register[i](value),
                self.flag[i](0)
            )
        fsm.goto_next()

    def write_flag(self, fsm, addr, value, resetvalue=0):
        state_cond = fsm.state == fsm.current
        for i, r in enumerate(self.register):
            self.seq.If(state_cond, addr == i)(
                self.register[i](value),
                self.flag[i](1),
                self.resetval[i](resetvalue)
            )
        fsm.goto_next()

    def wait(self, fsm, addr, value, polarity=True):
        if isinstance(addr, int):
            rval = self.register[addr]
        elif isinstance(addr, vtypes.Int):
            rval = self.register[addr.value]
        else:
            pat = [(addr == i, r) for i, r in enumerate(self.register)]
            pat.append((None, vtypes.IntX()))
            rval = vtypes.PatternMux(pat)

        if polarity:
            wait_cond = (rval == value)
        else:
            wait_cond = (rval != value)

        fsm.If(wait_cond).goto_next()

    def wait_flag(self, fsm, addr, value, resetvalue=0, polarity=True):
        if isinstance(addr, int):
            rval = self.register[addr]
        elif isinstance(addr, vtypes.Int):
            rval = self.register[addr.value]
        else:
            pat = [(addr == i, r) for i, r in enumerate(self.register)]
            pat.append((None, vtypes.IntX()))
            rval = vtypes.PatternMux(pat)

        if polarity:
            wait_cond = (rval == value)
        else:
            wait_cond = (rval != value)

        state_cond = fsm.state == fsm.current

        # flag reset
        for i, r in enumerate(self.register):
            self.seq.If(wait_cond, state_cond, addr == i)(
                self.register[i](resetvalue)
            )

        fsm.If(wait_cond).goto_next()


class AXISLiteRegister(AXISLite):
    __intrinsics__ = ('read', 'write', 'write_flag', 'wait',
                      'wait_flag') + _MutexFunction.__intrinsics__

    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=32,
                 noio=False, length=4, fsm_as_module=False):

        AXISLite.__init__(self, m, name, clk, rst, datawidth, addrwidth,
                          noio)

        self.fsm_as_module = fsm_as_module

        if not isinstance(length, int):
            raise TypeError("length must be 'int', not '%s'" %
                            str(type(length)))

        self.register = [self.m.Reg('_'.join(['', self.name, 'register', '%d' % i]),
                                    width=self.datawidth, initval=0, signed=True)
                         for i in range(length)]
        self.flag = [self.m.Reg('_'.join(['', self.name, 'flag', '%d' % i]), initval=0)
                     for i in range(length)]
        self.resetval = [self.m.Reg('_'.join(['', self.name, 'resetval', '%d' % i]),
                                    width=self.datawidth, initval=0, signed=True)
                         for i in range(length)]
        self.length = length
        self.maskwidth = self.m.Localparam('_'.join(['', self.name, 'maskwidth']),
                                           util.log2(length))
        self.mask = self.m.Localparam('_'.join(['', self.name, 'mask']),
                                      vtypes.Repeat(vtypes.Int(1, 1), self.maskwidth))
        self.shift = self.m.Localparam('_'.join(['', self.name, 'shift']),
                                       util.log2(self.datawidth // 8))

        self._set_register_lite_fsm()

    def _set_register_lite_fsm(self):
        fsm = FSM(self.m, '_'.join(['', self.name, 'register_fsm']),
                  self.clk, self.rst, as_module=self.fsm_as_module)

        # request
        addr, readvalid, writevalid = self.pull_request(cond=fsm)

        maskaddr = self.m.TmpReg(self.maskwidth)
        fsm.If(vtypes.Ors(readvalid, writevalid))(
            maskaddr((addr >> self.shift) & self.mask),
        )

        init_state = fsm.current

        # read
        read_state = fsm.current + 1
        fsm.If(readvalid).goto_from(init_state, read_state)
        fsm.set_index(read_state)

        rdata = self.m.TmpWire(self.datawidth, signed=True)
        pat = [(maskaddr == i, r) for i, r in enumerate(self.register)]
        pat.append((None, vtypes.IntX()))
        rval = vtypes.PatternMux(pat)
        rdata.assign(rval)

        flag = self.m.TmpWire()
        pat = [(maskaddr == i, r) for i, r in enumerate(self.flag)]
        pat.append((None, vtypes.IntX()))
        rval = vtypes.PatternMux(pat)
        flag.assign(rval)

        resetval = self.m.TmpWire(self.datawidth, signed=True)
        pat = [(maskaddr == i, r) for i, r in enumerate(self.resetval)]
        pat.append((None, vtypes.IntX()))
        rval = vtypes.PatternMux(pat)
        resetval.assign(rval)

        ack, valid = self.push_read_data(rdata, cond=fsm)

        # flag reset
        state_cond = fsm.state == fsm.current
        for i, r in enumerate(self.register):
            self.seq.If(state_cond, ack, flag, maskaddr == i)(
                self.register[i](resetval),
                self.flag[i](0)
            )

        fsm.If(valid).goto_init()
        fsm.If(ack, vtypes.Not(valid)).goto_next()
        fsm.If(valid).goto_init()

        # write
        write_state = fsm.current + 1
        fsm.If(writevalid).goto_from(init_state, write_state)
        fsm.set_index(write_state)

        data, mask, valid = self.pull_write_data(cond=fsm)

        state_cond = fsm.state == fsm.current
        for i, r in enumerate(self.register):
            self.seq.If(state_cond, valid, maskaddr == i)(
                self.register[i](data)
            )

        fsm.If(valid).goto_init()

    def read(self, fsm, addr):
        if isinstance(addr, int):
            rval = self.register[addr]
        elif isinstance(addr, vtypes.Int):
            rval = self.register[addr.value]
        else:
            pat = [(addr == i, r) for i, r in enumerate(self.register)]
            pat.append((None, vtypes.IntX()))
            rval = vtypes.PatternMux(pat)
        return rval

    def write(self, fsm, addr, value):
        state_cond = fsm.state == fsm.current
        for i, r in enumerate(self.register):
            self.seq.If(state_cond, addr == i)(
                self.register[i](value),
                self.flag[i](0)
            )
        fsm.goto_next()

    def write_flag(self, fsm, addr, value, resetvalue=0):
        state_cond = fsm.state == fsm.current
        for i, r in enumerate(self.register):
            self.seq.If(state_cond, addr == i)(
                self.register[i](value),
                self.flag[i](1),
                self.resetval[i](resetvalue)
            )
        fsm.goto_next()

    def wait(self, fsm, addr, value, polarity=True):
        if isinstance(addr, int):
            rval = self.register[addr]
        elif isinstance(addr, vtypes.Int):
            rval = self.register[addr.value]
        else:
            pat = [(addr == i, r) for i, r in enumerate(self.register)]
            pat.append((None, vtypes.IntX()))
            rval = vtypes.PatternMux(pat)

        if polarity:
            wait_cond = (rval == value)
        else:
            wait_cond = (rval != value)

        fsm.If(wait_cond).goto_next()

    def wait_flag(self, fsm, addr, value, resetvalue=0, polarity=True):
        if isinstance(addr, int):
            rval = self.register[addr]
        elif isinstance(addr, vtypes.Int):
            rval = self.register[addr.value]
        else:
            pat = [(addr == i, r) for i, r in enumerate(self.register)]
            pat.append((None, vtypes.IntX()))
            rval = vtypes.PatternMux(pat)

        if polarity:
            wait_cond = (rval == value)
        else:
            wait_cond = (rval != value)

        state_cond = fsm.state == fsm.current

        # flag reset
        for i, r in enumerate(self.register):
            self.seq.If(wait_cond, state_cond, addr == i)(
                self.register[i](resetvalue)
            )

        fsm.If(wait_cond).goto_next()



from __future__ import absolute_import
from __future__ import print_function
import os
import sys

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *

def mkChatterClear(length=1024):
    m = Module("chatter_clear")

    length = m.Parameter('length', length)
    
    clk = m.Input('CLK')
    rst = m.Input('RST')

    din = m.Input('din')
    dout = m.OutputReg('dout', initval=0)

    seq = Seq(m, 'seq', clk, rst)

    count = m.TmpReg(32)
    
    seq.If(din==dout)(
        count(0)
    )
    seq.If(din!=dout)(
        count.inc()
    )
    
    seq.If(count==length)(
        count(0)
    )
    seq.If(count==length)(
        dout(din)
    )

    seq.make_always()

    return m

def mkTest(length=1024):
    m = Module('test')
    
    main = mkChatterClear(length)
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']
    din = ports['din']
    dout = ports['dout']

    fsm = FSM(m, 'fsm', clk, rst)
    count = m.TmpReg(32, initval=0)

    fsm( din(0) )
    fsm( count.inc() )
    fsm.If(count==2000)( count(0) )
    fsm.goto_next(count==2000)

    fsm( din(1) )
    fsm( count.inc() )
    fsm.If(count==10)( count(0) )
    fsm.goto_next(count==10)

    fsm( din(0) )
    fsm( count.inc() )
    fsm.If(count==10)( count(0) )
    fsm.goto_next(count==10)

    fsm( din(1) )
    fsm( count.inc() )
    fsm.If(count==2000)( count(0) )
    fsm.goto_next(count==2000)

    fsm( din(0) )
    fsm( count.inc() )
    fsm.If(count==10)( count(0) )
    fsm.goto_next(count==10)

    fsm( din(1) )
    fsm( count.inc() )
    fsm.If(count==10)( count(0) )
    fsm.goto_next(count==10)

    fsm( din(0) )
    fsm( count.inc() )
    fsm.If(count==2000)( count(0) )
    fsm.goto_next(count==2000)

    fsm.make_always()
    
    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    nclk = simulation.next_clock
    
    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m

if __name__ == '__main__':
    #main = mkChatterClear()
    #verilog = main.to_verilog('tmp.v')
    #print(verilog)
    #exit()
    
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    # run simulator (Icarus Verilog)
    sim = simulation.Simulator(test)
    rslt = sim.run() # display=False
    #rslt = sim.run(display=True)
    print(rslt)

    # launch waveform viewer (GTKwave)
    #sim.view_waveform() # background=False
    #sim.view_waveform(background=True)



from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import copy
from collections import OrderedDict

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module


def get_width(node):
    ret = vtypes.get_width(node)
    if isinstance(node, int) and ret < 32:
        return 32
    if ret is None or isinstance(node, vtypes.AnyType):
        return 32
    return ret


def check_constant(obj):
    if not isinstance(obj, vtypes.VeriloggenNode):
        return True
    return False


def check_overwrite(obj):
    if check_constant(obj):
        return True
    if isinstance(obj, (tuple, list)):
        return True
    return False


class _Visitor(object):

    def generic_visit(self, node):
        raise TypeError("Type %s is not supported." % str(type(node)))

    def visit(self, node):
        visitor = getattr(
            self, 'visit_' + node.__class__.__name__, self.generic_visit)
        return visitor(node)


class _CachedVisitor(_Visitor):

    def __init__(self):
        _Visitor.__init__(self)
        self.visited_node = {}

    def visit(self, node):
        # check the cache
        if isinstance(node, (tuple, list)):
            return self._visit(node)

        if node in self.visited_node:
            return self.visited_node[node]

        ret = self._visit(node)
        self.visited_node[node] = ret
        return ret

    def _visit(self, node):
        visitor = getattr(
            self, 'visit_' + node.__class__.__name__, self.generic_visit)
        return visitor(node)


class _CommonVisitor(_CachedVisitor):

    def generic_visit(self, node):
        if isinstance(node, vtypes._Variable):
            return self.visit__Variable(node)

        if isinstance(node, vtypes._Constant):
            return self.visit__Constant(node)

        if isinstance(node, vtypes._BinaryOperator):
            return self.visit__BinaryOperator(node)

        if isinstance(node, vtypes._UnaryOperator):
            return self.visit__UnaryOperator(node)

        #raise TypeError("Type %s is not supported." % str(type(node)))
        return node

    def visit__Variable(self, node):
        return node

    def visit__Constant(self, node):
        return node.value

    def visit__BinaryOperator(self, node):
        return node

    def visit__UnaryOperator(self, node):
        return node


class ConstantVisitor(_CommonVisitor):

    def __init__(self, const_dict):
        _CommonVisitor.__init__(self)
        self.const_dict = const_dict

    # -------------------------------------------------------------------------
    def get_const(self, name):
        if name not in self.const_dict:
            raise KeyError("No such constant: '%s'" % name)
        return self.const_dict[name]

    def has_const(self, name):
        return (name in self.const_dict)

    def update_const(self, name, value):
        self.const_dict[name] = value

    def get_const_dict(self):
        return self.const_dict

    # -------------------------------------------------------------------------
    def _visit_param(self, node):
        if self.has_const(node.name):
            value = self.get_const(node.name)
            return value
        # unresolved node
        return node

    def visit_Parameter(self, node):
        return self._visit_param(node)

    def visit_Localparam(self, node):
        return self._visit_param(node)

    def visit_AnyType(self, node):
        return self._visit_param(node)

    # -------------------------------------------------------------------------
    def visit__BinaryOperator(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        lwidth = self.visit(get_width(node.left))
        rwidth = self.visit(get_width(node.right))
        if (check_constant(left) and
            check_constant(right) and
            check_constant(lwidth) and
                check_constant(rwidth)):
            return node.op(left, right, lwidth, rwidth)
        return node

    def visit__UnaryOperator(self, node):
        right = self.visit(node.right)
        rwidth = self.visit(get_width(node.right))
        if (check_constant(right) and
                check_constant(rwidth)):
            return node.op(right, rwidth)
        return node

    # -------------------------------------------------------------------------
    def visit_Pointer(self, node):
        var = self.visit(node.var)
        pos = self.visit(node.pos)
        if (check_constant(var) and
                check_constant(pos)):
            return node.op(var, pos)
        return node

    def visit_Slice(self, node):
        var = self.visit(node.var)
        msb = self.visit(node.msb)
        lsb = self.visit(node.lsb)
        if (check_constant(var) and
            check_constant(msb) and
                check_constant(lsb)):
            return node.op(var, msb, lsb)
        return node

    def visit_Cat(self, node):
        vars = [self.visit(var) for var in node.vars]
        widths = [self.visit(get_width(var)) for var in node.vars]
        for var, width in zip(vars, widths):
            if not check_constant(var):
                return node
            if not check_constant(width):
                return node
        return node.op(vars, widths)

    def visit_Repeat(self, node):
        var = self.visit(node.var)
        width = self.visit(get_width(node.var))
        times = self.visit(node.times)
        if (check_constant(var) and
            check_constant(width) and
                check_constant(times)):
            return node.op(var, width, times)
        return node

    def visit_Cond(self, node):
        condition = self.visit(node.condition)
        true_value = self.visit(node.true_value)
        false_value = self.visit(node.false_value)
        if (check_constant(condition) and
            check_constant(true_value) and
                check_constant(false_value)):
            return node.op(condition, true_value, false_value)
        return node


class ReplaceVisitor(ConstantVisitor):

    def visit_tuple(self, node):
        return tuple([self.visit(n) for n in node])

    def visit_list(self, node):
        return [self.visit(n) for n in node]

    # -------------------------------------------------------------------------
    def visit__BinaryOperator(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        lwidth = self.visit(get_width(node.left))
        rwidth = self.visit(get_width(node.right))
        if (check_constant(left) and
            check_constant(right) and
            check_constant(lwidth) and
                check_constant(rwidth)):
            return node.op(left, right, lwidth, rwidth)
        if check_constant(left):
            node.left = left
        if check_constant(right):
            node.right = right
        return node

    def visit__UnaryOperator(self, node):
        right = self.visit(node.right)
        rwidth = self.visit(get_width(node.right))
        if (check_constant(right) and
                check_constant(rwidth)):
            return node.op(right, rwidth)
        if check_constant(right):
            node.right = right
        return node

    def visit_Pointer(self, node):
        var = self.visit(node.var)
        pos = self.visit(node.pos)
        if (check_constant(var) and
                check_constant(pos)):
            return node.op(var, pos)
        if check_constant(var):
            node.var = var
        if check_constant(pos):
            node.pos = pos
        return node

    def visit_Slice(self, node):
        var = self.visit(node.var)
        msb = self.visit(node.msb)
        lsb = self.visit(node.lsb)
        if (check_constant(var) and
            check_constant(msb) and
                check_constant(lsb)):
            return node.op(var, msb, lsb)
        if check_constant(var):
            node.var = var
        if check_constant(msb):
            node.msb = msb
        if check_constant(lsb):
            node.lsb = lsb
        return node

    def visit_Cat(self, node):
        vars = [self.visit(var) for var in node.vars]
        widths = [self.visit(get_width(var)) for var in node.vars]
        node.vars = vars
        for var, width in zip(vars, widths):
            if not check_constant(var):
                return node
            if not check_constant(width):
                return node
        return node.op(vars, widths)

    def visit_Repeat(self, node):
        var = self.visit(node.var)
        width = self.visit(get_width(node.var))
        times = self.visit(node.times)
        if (check_constant(var) and
            check_constant(width) and
                check_constant(times)):
            return node.op(var, width, times)
        if check_constant(var):
            node.var = var
        if check_constant(times):
            node.times = times
        return node

    def visit_Cond(self, node):
        condition = self.visit(node.condition)
        true_value = self.visit(node.true_value)
        false_value = self.visit(node.false_value)
        if (check_constant(condition) and
            check_constant(true_value) and
                check_constant(false_value)):
            return node.op(condition, true_value, false_value)
        if check_constant(condition):
            node.condition = condition
        if check_constant(true_value):
            node.true_value = true_value
        if check_constant(false_value):
            node.false_value = false_value
        return node

    # -------------------------------------------------------------------------
    def visit_Posedge(self, node):
        name = self.visit(node.name)
        if check_overwrite(name):
            node.name = name
        return node

    def visit_Negedge(self, node):
        name = self.visit(node.name)
        if check_overwrite(name):
            node.name = name
        return node

    def visit_SensitiveAll(self, node):
        return node

    # -------------------------------------------------------------------------
    def visit_Subst(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        ldelay = self.visit(node.ldelay) if node.ldelay is not None else None
        rdelay = self.visit(node.rdelay) if node.rdelay is not None else None
        node.left = left
        if check_constant(right):
            node.right = right
        if check_constant(ldelay):
            node.ldelay = ldelay
        if check_constant(rdelay):
            node.rdelay = rdelay
        return node

    # -------------------------------------------------------------------------
    def visit_If(self, node):
        condition = self.visit(node.condition)
        true_statement = self.visit(node.true_statement)
        if node.false_statement is not None:
            false_statement = self.visit(node.false_statement)

        if check_overwrite(condition):
            node.condition = condition

        node.true_statement = true_statement
        if node.false_statement is not None:
            node.false_statement = false_statement

        return node

    def visit_For(self, node):
        pre = self.visit(node.pre)
        condition = self.visit(node.condition)
        post = self.visit(node.post)

        if check_overwrite(condition):
            node.condition = condition

        new_statement = []
        const_dict = copy.deepcopy(self.const_dict)

        # pre
        dst = pre.left.name
        val = pre.right
        const_dict[dst] = val

        while True:
            # cond
            cvisitor = ForReplaceVisitor(const_dict)
            cond = cvisitor.visit(condition)

            if not check_constant(cond):
                return node

            if not cond:
                break

            rslt = cvisitor.visit(copy.deepcopy(node.statement))
            new_statement.append(rslt)

            # post
            if post.left.name != dst:
                raise ValueError(
                    "Destination of 'post' must be same with 'pre': %s" % dst)

            const_dict[dst] = cvisitor.visit(post.right)

        return new_statement

    def visit_While(self, node):
        condition = self.visit(node.condition)
        statement = self.visit(node.statement)
        if check_constant(condition):
            node.condition = condition
        node.statement = statement
        return node

    def visit_Case(self, node):
        comp = self.visit(node.comp)
        statement = [self.visit(s) for s in node.statement]
        if check_constant(comp):
            node.comp = comp
        node.statement = statement
        return node

    def visit_Casex(self, node):
        comp = self.visit(node.comp)
        statement = [self.visit(s) for s in node.statement]
        if check_constant(comp):
            node.comp = comp
        node.statement = statement
        return node

    def visit_When(self, node):
        condition = self.visit(node.condition)
        statement = self.visit(node.statement)
        if check_constant(condition):
            node.condition = condition
        node.statement = statement
        return node

    def visit_ScopeIndex(self, node):
        self.visit(node.name)
        index = self.visit(node.index)
        if check_constant(index):
            node.index = index
        return node

    def visit_Scope(self, node):
        args = [self.visit(arg) for arg in node.args]
        node.args = args
        return node

    def visit_SystemTask(self, node):
        args = [self.visit(a) for a in node.args]
        node.args = args
        return node

    def visit_Event(self, node):
        sensitivity = [self.visit(s) for s in node.sensitivity]
        node.sensitivity = sensitivity
        return node

    def visit_Forever(self, node):
        statement = [self.visit(s) for s in node.statement]
        node.statement = statement
        return node

    def visit_Delay(self, node):
        delay = self.visit(node.delay)
        if check_constant(delay):
            node.delay = delay
        return node

    def visit_SingleStatement(self, node):
        statement = self.visit(node.statement)
        node.statement = statement
        return node

    def visit_Function(self, node):
        # not implemented
        return node

    def visit_FunctionCall(self, node):
        self.visit(node.func)
        args = [self.visit(a) for a in node.args]
        node.args = args
        return node

    def visit_Task(self, node):
        # not implemented
        return node

    def visit_TaskCall(self, node):
        self.visit(node.func)
        args = [self.visit(a) for a in node.args]
        node.args = args
        return node


class ForReplaceVisitor(ReplaceVisitor):

    def visit__Variable(self, node):
        return self._visit_param(node)


class ModuleReplaceVisitor(_CachedVisitor):
    tmp_count = 0

    def __init__(self, mod, const_dict=None, submodule=None):
        _CachedVisitor.__init__(self)

        if not isinstance(mod, module.Module) and not isinstance(mod, module.Generate):
            raise TypeError("Not supported object type: '%s'", str(type(mod)))

        if const_dict is None:
            const_dict = OrderedDict()

        if submodule is None:
            submodule = OrderedDict()

        self.mod = mod
        self.const_dict = const_dict
        self.submodule = submodule

        self.replace_visitor = None

    def resolve(self):
        self._update_const_dict()
        self.replace_visitor = ReplaceVisitor(self.const_dict)
        return self.visit(self.mod)

    def _update_const_dict(self):
        params = self.mod.get_params()
        localparams = self.mod.get_localparams()

        unresolved = OrderedDict()
        unresolved.update(params)
        unresolved.update(localparams)

        prev_size = len(unresolved) + 1

        while unresolved and len(unresolved) < prev_size:
            prev_size = len(unresolved)
            unresolved = self._update_const_dict_iter(unresolved)

        if unresolved:
            varlist = ', '.join(unresolved.keys())
            raise ValueError(
                'Not all parameters could be resolved: %s' % varlist)

    def _update_const_dict_iter(self, unresolved):
        const_visitor = ConstantVisitor(self.const_dict)
        next_unresolved = OrderedDict()

        for name, param in unresolved.items():
            value = const_visitor.visit(param.value)

            if not check_constant(value):
                next_unresolved[name] = param
                continue

            if name not in self.const_dict:
                const_visitor.update_const(name, value)

        self.const_dict = const_visitor.get_const_dict()

        return next_unresolved

    def visit_Module(self, node):
        params = node.get_params()
        for param in params.values():
            if param.width is not None:
                param.width = self.replace_visitor.visit(param.width)
            param.value = self.replace_visitor.visit(param.value)

        localparams = node.get_localparams()
        for localparam in localparams.values():
            if localparam.width is not None:
                localparam.width = self.replace_visitor.visit(localparam.width)
            localparam.value = self.replace_visitor.visit(localparam.value)

        ports = node.get_ports()
        for port in ports.values():
            if port.width is not None:
                port.width = self.replace_visitor.visit(port.width)

        vars = node.get_vars()
        for var in vars.values():
            if var.width is not None:
                var.width = self.replace_visitor.visit(var.width)
            if var.dims is not None:
                var.dims = self.replace_visitor.visit(var.dims)

        for asg in node.assign:
            self.visit(asg)

        for alw in node.always:
            self.visit(alw)

        for ini in node.initial:
            self.visit(ini)

        for ins in node.instance.values():
            self.visit(ins)

        for gen in node.generate.values():
            self.visit(gen)

        return node

    def visit__Variable(self, node):
        width = self.replace_visitor.visit(node.width) if node.width is not None else None
        dims = self.replace_visitor.visit(node.dims) if node.dims is not None else None
        initval = self.replace_visitor.visit(node.initval) if node.initval is not None else None
        if check_constant(width):
            node.width = width
        if node.dims is not None:
            new_dims = []
            for dim, old_dim in zip(dims, node.dims):
                if check_constant(dim):
                    new_dims.append(dim)
                else:
                    new_dims.append(old_dim)
            node.dims = tuple(dims)
        if check_constant(initval):
            node.initval = initval
        return node

    def visit_Always(self, node):
        sensitivity = self.replace_visitor.visit(node.sensitivity)
        self.replace_visitor.visit(node.statement)
        if check_constant(sensitivity):
            node.sensitivity = sensitivity
        return node

    def visit_Assign(self, node):
        self.replace_visitor.visit(node.statement)
        return node

    def visit_Initial(self, node):
        for s in node.statement:
            self.replace_visitor.visit(s)
        return node

    def visit_Instance(self, node):
        params = [(k, self.replace_visitor.visit(v)) for k, v in node.params]
        ports = [(k, self.replace_visitor.visit(v)) for k, v in node.ports]
        node.params = params
        node.ports = ports

        inst_const_dict = OrderedDict()
        for k, v in params:
            inst_const_dict[k] = v

        if isinstance(node.module, (module.StubModule, str)):
            return node

        mod = copy.deepcopy(node.module)
        prev_name = mod.name
        new_name = prev_name

        while True:
            if new_name not in self.submodule:
                self.submodule[new_name] = mod
                self.mod.submodule[new_name] = mod
                mod.name = new_name
                break

            new_name = '_'.join(
                [prev_name, 'copy', str(ModuleReplaceVisitor.tmp_count)])
            ModuleReplaceVisitor.tmp_count += 1

        mvisitor = ModuleReplaceVisitor(mod, inst_const_dict, self.submodule)
        mod = mvisitor.resolve()
        node.module = mod

        return node

    def visit_GenerateFor(self, node):
        # raise NotImplementedError(
        #    "GenerateFor statement is not currently supported.")
        return node

    def visit_GenerateIf(self, node):
        # raise NotImplementedError(
        #    "GenerateIf statement is not currently supported.")
        return node


def resolve(m, const_dict=None):
    mvisitor = ModuleReplaceVisitor(m, const_dict)
    return mvisitor.resolve()



from __future__ import absolute_import
from __future__ import print_function

from .axim import AXIM, AXIMLite, AXIMVerify, AXIMLiteVerify
from .axis import AXIS, AXISLite, AXISRegister, AXISLiteRegister
from .axistreamin import AXIStreamIn, AXIStreamInFifo
from .axistreamout import AXIStreamOut, AXIStreamOutFifo
from .axim_for_axistream import AXIM_for_AXIStreamIn, AXIM_for_AXIStreamOut, AXIM2



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
from veriloggen.thread import RAM
import veriloggen.dataflow as dataflow


def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    addrwidth = int(math.log(n, 2)) * 2
    myram = RAM(m, 'myram', clk, rst, datawidth, addrwidth, 2)
    myram.disable_write(1)

    df = dataflow.DataflowManager(m, clk, rst)
    fsm = FSM(m, 'fsm', clk, rst)

    fsm.goto_next()

    # dataflow
    value = df.Counter(size=64)
    value = value - 1

    # write dataflow (Dataflow -> RAM)
    wport = 0
    waddr = 0
    wlen = 64
    done = myram.write_dataflow(wport, waddr, value, wlen, cond=fsm)
    fsm.goto_next()
    fsm.If(done).goto_next()

    # verify
    sum = m.Reg('sum', 32, initval=0)
    expected_sum = (waddr + waddr + wlen - 1) * wlen // 2 - wlen

    seq = Seq(m, 'seq', clk, rst)
    seq.If(myram[0].wenable)(
        sum.add(myram[0].wdata)
    )
    seq.Then().If(myram[0].addr == wlen - 1).Delay(2)(
        Systask('display', 'sum=%d expected_sum=%d', sum, expected_sum),
        If(NotEql(sum, expected_sum))(Display('# verify: FAILED')).Else(Display('# verify: PASSED'))
    )

    return m


def mkTest(memimg_name=None):
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    # simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)
    lines = rslt.splitlines()
    if simtype == 'verilator' and lines[-1].startswith('-'):
        rslt = '\n'.join(lines[:-1])
    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import ram_rtl

expected_verilog = """
module test;

  reg CLK;
  reg RST;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST
);

  wire [14-1:0] myram_0_addr;
  wire [32-1:0] myram_0_rdata;
  wire [32-1:0] myram_0_wdata;
  wire myram_0_wenable;
  wire myram_0_enable;

  myram
  inst_myram
  (
    .CLK(CLK),
    .myram_0_addr(myram_0_addr),
    .myram_0_rdata(myram_0_rdata),
    .myram_0_wdata(myram_0_wdata),
    .myram_0_wenable(myram_0_wenable),
    .myram_0_enable(myram_0_enable)
  );

  reg [32-1:0] count;
  reg [32-1:0] sum;
  reg [32-1:0] addr;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  assign myram_0_wdata = (fsm == 1)? count : 'hx;
  assign myram_0_wenable = (fsm == 1)? 1'd1 : 0;
  assign myram_0_addr = (fsm == 2)? addr : 
                        (fsm == 1)? addr : 'hx;
  assign myram_0_enable = (fsm == 2)? 1'd1 : 
                          (fsm == 1)? 1'd1 : 0;
  localparam _tmp_0 = 1;
  wire [_tmp_0-1:0] _tmp_1;
  assign _tmp_1 = fsm == 2;
  reg [_tmp_0-1:0] __tmp_1_1;
  reg [32-1:0] _d1_fsm;
  reg _fsm_cond_2_0_1;
  reg _fsm_cond_3_1_1;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      _d1_fsm <= fsm_init;
      addr <= 0;
      count <= 0;
      sum <= 0;
      _fsm_cond_2_0_1 <= 0;
      _fsm_cond_3_1_1 <= 0;
    end else begin
      _d1_fsm <= fsm;
      case(_d1_fsm)
        fsm_2: begin
          if(_fsm_cond_2_0_1) begin
            $display("sum=%d", sum);
          end 
        end
        fsm_3: begin
          if(_fsm_cond_3_1_1) begin
            $display("sum=%d", sum);
          end 
        end
      endcase
      case(fsm)
        fsm_init: begin
          addr <= 0;
          count <= 0;
          sum <= 0;
          fsm <= fsm_1;
        end
        fsm_1: begin
          addr <= addr + 1;
          count <= count + 1;
          if(count == 15) begin
            addr <= 0;
            count <= 0;
          end 
          if(count == 15) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          addr <= addr + 1;
          count <= count + 1;
          if(__tmp_1_1) begin
            sum <= sum + myram_0_rdata;
          end 
          _fsm_cond_2_0_1 <= __tmp_1_1;
          if(count == 15) begin
            addr <= 0;
            count <= 0;
          end 
          if(count == 15) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(__tmp_1_1) begin
            sum <= sum + myram_0_rdata;
          end 
          _fsm_cond_3_1_1 <= __tmp_1_1;
          fsm <= fsm_4;
        end
        fsm_4: begin
          $display("expected_sum=%d", 120);
          fsm <= fsm_5;
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      __tmp_1_1 <= 0;
    end else begin
      __tmp_1_1 <= _tmp_1;
    end
  end


endmodule



module myram
(
  input CLK,
  input [14-1:0] myram_0_addr,
  output [32-1:0] myram_0_rdata,
  input [32-1:0] myram_0_wdata,
  input myram_0_wenable,
  input myram_0_enable
);

  reg [32-1:0] myram_0_rdata_out;
  assign myram_0_rdata = myram_0_rdata_out;
  reg [32-1:0] mem [0:16384-1];

  always @(posedge CLK) begin
    if(myram_0_enable) begin
      if(myram_0_wenable) begin
        mem[myram_0_addr] <= myram_0_wdata;
        myram_0_rdata_out <= myram_0_wdata;
      end else begin
        myram_0_rdata_out <= mem[myram_0_addr];
      end
    end 
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = ram_rtl.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.resolver.resolver as resolver

def mkBram(datawidth=32, addrwidth=10, numports=2):
    m = Module('BRAM%d' % numports)

    datawidth = m.Parameter('DATA_WIDTH', datawidth)
    addrwidth = m.Parameter('ADDR_WIDTH', addrwidth)

    clk = m.Input('CLK')
    ports = []
    for i in range(numports):
        addr = m.Input('ADDR%d' % i, addrwidth)
        din = m.Input('DIN%d' % i, datawidth)
        we = m.Input('WE%d' % i)
        dout = m.Output('DOUT%d' % i, datawidth)
        delay_addr = m.Reg('delay_ADDR%d' % i, addrwidth)
        ports.append( (addr, din, we, dout, delay_addr) )

    mem = m.Reg('mem', datawidth, Int(2)**addrwidth)

    for i in range(numports):
        addr, din ,we, dout, delay_addr = ports[i]
        m.Always(Posedge(clk))(
            If(we)(
                mem[addr](din)
            ),
            delay_addr(addr)
        )
        m.Assign(dout(mem[delay_addr]))

    return m

def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    addrwidth = int(math.log(n, 2)) * 2

    datawidth = m.Parameter('DATA_WIDTH', datawidth)
    addrwidth = m.Parameter('ADDR_WIDTH', addrwidth)
    
    bram = mkBram(datawidth, addrwidth, numports)

    bram_ports = []
    for i in range(numports):
        addr = m.Reg('bram_addr%d' % i, addrwidth)
        din = m.Reg('bram_din%d' % i, datawidth)
        we = m.Reg('bram_we%d' % i)
        dout = m.Wire('bram_dout%d' % i, datawidth)
        bram_ports.append( [addr, din, we, dout] )

    params = [ (datawidth.name, datawidth), 
               (addrwidth.name, addrwidth) ]
    ports = [ clk ]
    for bram_port in bram_ports:
        ports.extend(bram_port)

    # BRAM instance
    m.Instance(bram, 'bram_inst', params=params, ports=ports)

    # example how to access BRAM
    fsm = FSM(m, 'fsm', clk, rst)
    
    for addr, din, we, dout in bram_ports:
        fsm.add( addr(0), din(0), we(0) )

    fsm.goto_next()

    count = m.Reg('count', 32, initval=0)
    for port, (addr, din, we, dout) in enumerate(bram_ports):
        fsm.add( addr(count+port*n//numports), din(count), we(1), count.inc() )

    fsm.goto_next(cond=count==n//numports-1)
    
    fsm.make_always()
    
    return m

def mkOrig():
    m = mkMain()
    m = resolver.resolve(m)
    return m
    
if __name__ == '__main__':
    orig = mkOrig()
    verilog = orig.to_verilog()
    print(verilog)
